#!/usr/bin/env python

"""
A native function library for a really simple virtual processor.

Copyright (C) 2007, 2008, 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.program import DataValue
import operator

class Library:

    "Native function implementations."

    def __init__(self, machine, constants):

        """
        Initialise the library with the 'machine' and the 'constants' addresses
        dictionary.
        """

        self.machine = machine
        self.constants = constants

        # Native class constants.

        cls = self.machine._get_class("__builtins__", "int")
        self.int_class = cls.location
        self.int_instance = cls.instance_template_location
        cls = self.machine._get_class("__builtins__", "list")
        if cls is not None:
            self.list_class = cls.location
            self.list_instance = cls.instance_template_location
        cls = self.machine._get_class("__builtins__", "IndexError")
        if cls is not None:
            self.index_error = cls.location
            self.index_error_instance = cls.instance_template_location
        cls = self.machine._get_class("__builtins__", "basestring")
        if cls is not None:
            self.str_class = cls.location
            self.str_instance = cls.instance_template_location

        self.tuple_class = self.machine.tuple_class
        self.type_error_instance = self.machine.type_error_instance

        self.frame_stack = self.machine.frame_stack
        self.local_sp_stack = self.machine.local_sp_stack

    def builtins_int_arithmetic_op(self, op):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_value = self.frame_stack[frame]
        right_value = self.frame_stack[frame + 1]

        # Test operand suitability.
        # NOTE: Support other types.

        if not (self.machine._CheckInstance(left_value.ref, self.int_class) and
            self.machine._CheckInstance(right_value.ref, self.int_class)):

            self.machine.exception = self.machine._MakeObject(2, self.type_error_instance)
            return self.machine.RaiseException()

        # NOTE: Assume single location for data.

        left_data = left_value.ref + 1
        right_data = right_value.ref + 1

        # Make a new object.

        addr = self.machine._MakeObject(2, self.int_instance)

        # Store the result.
        # NOTE: The data is considered ready to use.

        self.machine.save(addr + 1, op(self.machine.load(left_data), self.machine.load(right_data)))

        # Return the new object.
        # Introduce object as context for the new object.

        self.machine.result = DataValue(addr, addr)

    def builtins_logical_op(self, operand_class, op):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_value = self.frame_stack[frame]
        right_value = self.frame_stack[frame + 1]

        # Test operand suitability.
        # NOTE: Handle comparisons of incompatible types more appropriately.
        # NOTE: Return NotImplemented.

        if not (self.machine._CheckInstance(left_value.ref, operand_class) and
            self.machine._CheckInstance(right_value.ref, operand_class)):

            notimpl = self.constants[NotImplemented]
            self.machine.result = DataValue(notimpl, notimpl)
            return

        # NOTE: Assume single location for data.

        left_data = left_value.ref + 1
        right_data = right_value.ref + 1

        # Test the data.
        # NOTE: The data is considered ready to use.

        if op(self.machine.load(left_data), self.machine.load(right_data)):
            self.machine.result = DataValue(self.constants[True], self.constants[True])
        else:
            self.machine.result = DataValue(self.constants[False], self.constants[False])

    # Operators.
    # Although this takes a short-cut by using the operator module, testing is
    # still performed on the operands to ensure that they qualify for these
    # native operations.

    def builtins_int_add(self):
        return self.builtins_int_arithmetic_op(operator.add)

    def builtins_int_sub(self):
        return self.builtins_int_arithmetic_op(operator.sub)

    def builtins_int_pow(self):
        return self.builtins_int_arithmetic_op(operator.pow)

    def builtins_int_lt(self):
        return self.builtins_logical_op(self.int_class, operator.lt)

    def builtins_int_le(self):
        return self.builtins_logical_op(self.int_class, operator.le)

    def builtins_int_gt(self):
        return self.builtins_logical_op(self.int_class, operator.gt)

    def builtins_int_ge(self):
        return self.builtins_logical_op(self.int_class, operator.ge)

    def builtins_int_eq(self):
        return self.builtins_logical_op(self.int_class, operator.eq)

    def builtins_int_ne(self):
        return self.builtins_logical_op(self.int_class, operator.ne)

    def builtins_str_lt(self):
        return self.builtins_logical_op(self.str_class, operator.lt)

    def builtins_str_le(self):
        return self.builtins_logical_op(self.str_class, operator.le)

    def builtins_str_gt(self):
        return self.builtins_logical_op(self.str_class, operator.gt)

    def builtins_str_ge(self):
        return self.builtins_logical_op(self.str_class, operator.ge)

    def builtins_str_eq(self):
        return self.builtins_logical_op(self.str_class, operator.eq)

    def builtins_str_ne(self):
        return self.builtins_logical_op(self.str_class, operator.ne)

    def builtins_int_and(self):
        return self.builtins_int_arithmetic_op(operator.and_)

    def builtins_int_or(self):
        return self.builtins_int_arithmetic_op(operator.or_)

    # Specific operator methods.

    def builtins_int_bool(self):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_value = self.frame_stack[frame]

        # Test operand suitability.

        if not self.machine._CheckInstance(left_value.ref, self.int_class):
            self.machine.exception = self.machine._MakeObject(2, self.type_error_instance)
            return self.machine.RaiseException()

        # NOTE: Assume single location for data.

        left_data = left_value.ref + 1

        # Test the data.
        # NOTE: The data is considered ready to use.

        if self.machine.load(left_data) != 0:
            self.machine.result = DataValue(self.constants[True], self.constants[True])
        else:
            self.machine.result = DataValue(self.constants[False], self.constants[False])

    def builtins_int_neg(self):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_value = self.frame_stack[frame]

        # Test operand suitability.

        if not self.machine._CheckInstance(left_value.ref, self.int_class):
            self.machine.exception = self.machine._MakeObject(2, self.type_error_instance)
            return self.machine.RaiseException()

        # NOTE: Assume single location for data.

        left_data = left_value.ref + 1

        # Make a new object.

        addr = self.machine._MakeObject(2, self.int_instance)

        # Store the result.
        # NOTE: The data is considered ready to use.

        self.machine.save(addr + 1, -self.machine.load(left_data))

        # Return the new object.
        # Introduce object as context for the new object.

        self.machine.result = DataValue(addr, addr)

    # Various built-in methods.

    def builtins_bool_bool(self):
        frame = self.local_sp_stack[-1]

        # Get operands addresses.

        left_value = self.frame_stack[frame]
        self.machine.result = DataValue(left_value.ref, left_value.ref)

    def builtins_list_new(self):
        frame = self.local_sp_stack[-1]

        # The first parameter should be the instance.

        list_value = self.frame_stack[frame]

        # Make a new sequence.
        # NOTE: Using an arbitrary size.

        new_fragment = self.machine._MakeFragment(1, 5) # include the header

        # Complete the list instance by saving the fragment reference.
        # NOTE: This requires an attribute in the list structure.

        addr = list_value.ref + 1
        self.machine.save(addr, DataValue(None, new_fragment))

    def builtins_list_getitem(self):
        frame = self.local_sp_stack[-1]

        # Get the operand address.

        item_value = self.frame_stack[frame + 1]

        # Get the list address.

        obj_value = self.frame_stack[frame]

        # Get the fragment address.
        # NOTE: Assume single location for header.

        fragment = self.machine.load(obj_value.ref + 1)

        # Get the fragment header.

        header = self.machine.load(fragment.ref)
        nelements = header.occupied_size - 1

        # Test operand suitability.

        if self.machine._CheckInstance(item_value.ref, self.int_class):

            # NOTE: Assume single location for data and header.

            item_pos = self.machine.load(item_value.ref + 1)

            if item_pos >= 0 and item_pos < nelements:
                pass
            elif item_pos < 0 and item_pos >= -nelements:
                item_pos = nelements + item_pos
            else:
                self.machine.exception = self.machine._MakeObject(2, self.index_error_instance)
                return self.machine.RaiseException()

        else:
            self.machine.exception = self.machine._MakeObject(2, self.type_error_instance)
            return self.machine.RaiseException()

        # NOTE: Assume single location for header.

        self.machine.result = self.machine.load(fragment.ref + 1 + item_pos)

    def builtins_list_len(self):
        frame = self.local_sp_stack[-1]

        # Get the list address.

        obj_value = self.frame_stack[frame]

        # Get the fragment address.
        # NOTE: Assume single location for header.

        fragment = self.machine.load(obj_value.ref + 1)

        # Get the fragment header.

        header = self.machine.load(fragment.ref)
        nelements = header.occupied_size - 1

        # Make a new object.

        addr = self.machine._MakeObject(2, self.int_instance)

        # Store the result.
        # NOTE: The data is considered ready to use.

        self.machine.save(addr + 1, nelements)

        # Return the new object.
        # Introduce object as context for the new object.

        self.machine.result = DataValue(addr, addr)

    def builtins_list_append(self):
        frame = self.local_sp_stack[-1]

        # Get operand address.

        arg_value = self.frame_stack[frame + 1]

        # Get the list address.

        obj_value = self.frame_stack[frame]

        # Get the fragment address.
        # NOTE: Assume single location for header.

        fragment = self.machine.load(obj_value.ref + 1)

        # Get the fragment header.

        header = self.machine.load(fragment.ref)

        # Attempt to add the reference.

        if header.occupied_size < header.allocated_size:
            self.machine.save(fragment.ref + header.occupied_size, arg_value)
            header.occupied_size += 1
        else:

            # Make a new fragment, maintaining more space than currently
            # occupied in order to avoid reallocation.

            new_fragment = self.machine._MakeFragment(header.occupied_size + 1, header.occupied_size * 2)

            # Copy existing elements.

            for i in range(1, header.occupied_size):
                self.machine.save(new_fragment + i, self.machine.load(fragment.ref + i))

            self.machine.save(new_fragment + header.occupied_size, arg_value)

            # Set the new fragment in the object.
            # NOTE: The old fragment could be deallocated.

            self.machine.save(obj_value.ref + 1, DataValue(None, new_fragment))

    def builtins_tuple_len(self):
        frame = self.local_sp_stack[-1]

        # Get the tuple address.

        obj_value = self.frame_stack[frame]

        # Get the header.
        # NOTE: Assume single location for header.

        header = self.machine.load(obj_value.ref)
        nelements = header.size - 1

        # Make a new object.

        addr = self.machine._MakeObject(2, self.int_instance)

        # Store the result.
        # NOTE: The data is considered ready to use.

        self.machine.save(addr + 1, nelements)

        # Return the new object.
        # Introduce object as context for the new object.

        self.machine.result = DataValue(addr, addr)

    def builtins_tuple_getitem(self):
        frame = self.local_sp_stack[-1]

        # Get the operand address.

        item_value = self.frame_stack[frame + 1]

        # Get the tuple address.

        obj_value = self.frame_stack[frame]

        # Get the header.
        # NOTE: Assume single location for header.

        header = self.machine.load(obj_value.ref)
        nelements = header.size - 1

        # NOTE: Assume single location for data and header.

        item_pos = self.machine.load(item_value.ref + 1)

        if item_pos >= 0 and item_pos < nelements:
            pass
        elif item_pos < 0 and item_pos >= -nelements:
            item_pos = nelements + item_pos
        else:
            self.machine.exception = self.machine._MakeObject(2, self.index_error_instance)
            return self.machine.RaiseException()

        # NOTE: Assume single location for header.

        self.machine.result = self.machine.load(obj_value.ref + 1 + item_pos)

    def builtins_object_init(self):
        pass

    def builtins_isinstance(self):
        frame = self.local_sp_stack[-1]

        # Get the operand addresses.

        obj_value = self.frame_stack[frame]
        cls_value = self.frame_stack[frame + 1]

        if self.machine._CheckInstance(obj_value.ref, cls_value.ref):
            self.machine.result = DataValue(self.constants[True], self.constants[True])
        else:
            self.machine.result = DataValue(self.constants[False], self.constants[False])

    native_functions = {

        # Native method implementations:

        "__builtins__.int.__add__" : builtins_int_add,
        "__builtins__.int.__radd__" : builtins_int_add,               # NOTE: To be made distinct.
        "__builtins__.int.__sub__" : builtins_int_sub,
        "__builtins__.int.__pow__" : builtins_int_pow,
        "__builtins__.int.__iadd__" : builtins_int_add,
        "__builtins__.int.__bool__" : builtins_int_bool,
        "__builtins__.int.__neg__" : builtins_int_neg,
        "__builtins__.int.__lt__" : builtins_int_lt,
        "__builtins__.int.__le__" : builtins_int_le,
        "__builtins__.int.__gt__" : builtins_int_gt,
        "__builtins__.int.__ge__" : builtins_int_ge,
        "__builtins__.int.__eq__" : builtins_int_eq,
        "__builtins__.int.__ne__" : builtins_int_ne,
        "__builtins__.int.__and__" : builtins_int_and,
        "__builtins__.int.__rand__" : builtins_int_and,
        "__builtins__.int.__or__" : builtins_int_or,
        "__builtins__.int.__ror__" : builtins_int_or,
        "__builtins__.bool.__bool__" : builtins_bool_bool,
        "__builtins__.list.__getitem__" : builtins_list_getitem,
        "__builtins__.list.__len__" : builtins_list_len,
        "__builtins__.list.append" : builtins_list_append,
        "__builtins__.tuple.__len__" : builtins_tuple_len,
        "__builtins__.tuple.__getitem__" : builtins_tuple_getitem,
        "__builtins__.basestring.__lt__" : builtins_str_lt,
        "__builtins__.basestring.__le__" : builtins_str_le,
        "__builtins__.basestring.__gt__" : builtins_str_gt,
        "__builtins__.basestring.__ge__" : builtins_str_ge,
        "__builtins__.basestring.__eq__" : builtins_str_eq,
        "__builtins__.basestring.__ne__" : builtins_str_ne,

        # Native initialisers:

        "__builtins__.object.__init__" : builtins_object_init,        # NOTE: A no-operation.
        "__builtins__.BaseException.__init__" : builtins_object_init, # NOTE: To be made distinct, potentially in the builtins module.

        # Native instantiator helpers:

        "__builtins__.list.__new__" : builtins_list_new,

        # Native helper functions:

        "__builtins__._isinstance" : builtins_isinstance,
        }

# vim: tabstop=4 expandtab shiftwidth=4
